import datetime
import unittest

from sdx.common.survey import Survey


class SurveyTests(unittest.TestCase):

    def test_datetime_ms_with_colon_in_timezone(self):
        rv = Survey.parse_timestamp("2017-01-11T17:18:53.020222+00:00")
        self.assertIsInstance(rv, datetime.datetime)

    def test_datetime_ms_with_timezone(self):
        rv = Survey.parse_timestamp("2017-01-11T17:18:53.020222+0000")
        self.assertIsInstance(rv, datetime.datetime)

    def test_datetime_zulu(self):
        rv = Survey.parse_timestamp("2017-01-11T17:18:53Z")
        self.assertIsInstance(rv, datetime.datetime)

    def test_date_iso(self):
        rv = Survey.parse_timestamp("2017-01-11")
        self.assertNotIsInstance(rv, datetime.datetime)
        self.assertIsInstance(rv, datetime.date)

    def test_date_diary(self):
        rv = Survey.parse_timestamp("11/07/2017")
        self.assertNotIsInstance(rv, datetime.datetime)
        self.assertIsInstance(rv, datetime.date)

    def test_date_period(self):
        rv = Survey.parse_timestamp("201605")
        self.assertNotIsInstance(rv, datetime.datetime)
        self.assertIsInstance(rv, datetime.date)

    def test_load_survey(self):
        ids = Survey.identifiers({
            "survey_id": "134",
            "tx_id": "27923934-62de-475c-bc01-433c09fd38b8",
            "collection": {
                "instrument_id": "0005",
                "period": "201704"
            },
            "metadata": {
                "user_id": "123456789",
                "ru_ref": "12345678901A"
            }
        })
        rv = Survey.load_survey(ids, "sdx.common.test", "data/{survey_id}.{inst_id}.json")
        self.assertIsNotNone(rv)

    def test_load_survey_miss(self):
        ids = Survey.identifiers({
            "survey_id": "127",
            "tx_id": "27923934-62de-475c-bc01-433c09fd38b8",
            "collection": {
                "instrument_id": "0001",
                "period": "201704"
            },
            "metadata": {
                "user_id": "123456789",
                "ru_ref": "12345678901A"
            }
        })
        rv = Survey.load_survey(ids, "sdx.common.test", "data/{survey_id}.{inst_id}.json")
        self.assertIsNone(rv)



#!/usr/bin/env python
#   coding: UTF-8

import argparse
from io import BytesIO
import json
import os
import sys
import uuid

import arrow
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import SimpleDocTemplate, Table, TableStyle, Paragraph
from reportlab.platypus.flowables import HRFlowable
from reportlab.lib.enums import TA_LEFT, TA_CENTER

__doc__ = """
SDX PDF Transformer.

Example:

python transform/transformers/PDFTransformer.py --survey transform/surveys/144.0001.json \\
< tests/replies/ukis-01.json > output.pdf

"""
styles = getSampleStyleSheet()

# Basic text style
style_n = styles["BodyText"]
style_n.alignment = TA_LEFT
style_n.spaceAfter = 25

# Subheading style
style_sh = styles["Heading2"]
style_sh.alignment = TA_LEFT

# Sub-subheading style (questions)
style_ssh = styles["Heading3"]
style_ssh.alignment = TA_LEFT

# Main heading style
style_h = styles['Heading1']
style_h.alignment = TA_CENTER


# Answer Style
style_answer = ParagraphStyle(name='BodyText', parent=styles['Normal'], spaceBefore=6)
style_answer.alignment = TA_LEFT
style_answer.fontName = "Helvetica-Bold"
style_answer.textColor = colors.red
style_answer.fontSize = 15
style_answer.leading = 20
style_answer.spaceAfter = 20

MAX_ANSWER_CHARACTERS_PER_LINE = 35


class PDFTransformer(object):

    @staticmethod
    def render(survey, response):
        buffer = BytesIO()
        doc = SimpleDocTemplate(buffer, pagesize=A4)
        doc.build(PDFTransformer.get_elements(survey, response))

        pdf = buffer.getvalue()
        buffer.close()

        return pdf

    @staticmethod
    def render_to_file(survey, response):
        rndm_name = uuid.uuid4()

        os.makedirs("./tmp/%s" % rndm_name)

        tmp_name = "./tmp/%s/%s.pdf" % (rndm_name, rndm_name)
        doc = SimpleDocTemplate(tmp_name, pagesize=A4)
        doc.build(PDFTransformer.get_elements())

        return os.path.realpath(tmp_name)

    @staticmethod
    def get_elements(survey, response):

        elements = []
        table_style_data = [('TEXTCOLOR', (0, 0), (-1, -1), colors.black),
                            ('VALIGN', (0, 0), (-1, -1), 'TOP'),
                            ('LINEBELOW', (0, 0), (-1, -1), 1, colors.black),
                            ('BOX', (0, 0), (-1, -1), 1, colors.black),
                            ('BOX', (0, 0), (0, -1), 1, colors.black),
                            ('BACKGROUND', (0, 0), (1, 0), colors.lightblue)]

        table_style = TableStyle(table_style_data)
        table_style.spaceAfter = 25

        heading_style = TableStyle(table_style_data)
        heading_style.spaceAfter = 25
        heading_style.add('SPAN', (0, 0), (1, 0))
        heading_style.add('ALIGN', (0, 0), (1, 0), 'CENTER')

        localised_date_str = PDFTransformer.get_localised_date(response['submitted_at'])

        heading_data = [[Paragraph(survey['title'], style_h)]]
        heading_data.append(['Form Type', response['collection']['instrument_id']])
        heading_data.append(['Respondent', response['metadata']['ru_ref']])
        heading_data.append(['Submitted At', localised_date_str])

        heading = Table(heading_data, style=heading_style, colWidths='*')

        elements.append(heading)

        for question_group in filter(lambda x: 'title' in x, survey['question_groups']):

            section_heading = True

            for question in filter(lambda x: 'text' in x, question_group['questions']):
                if question['question_id'] in response['data']:
                    try:
                        answer = response['data'][question['question_id']]
                    except KeyError:
                        answer = ''

                    # Output the section header if we haven't already
                    # Checking here so that whole sections are suppressed
                    # if they have no answers.
                    if section_heading:
                        elements.append(HRFlowable(width="100%"))
                        elements.append(Paragraph(question_group['title'], style_sh))
                        section_heading = False

                    # Question not output if answer is empty
                    text = question.get("text")
                    if not text[0].isdigit():
                        text = " ".join((question.get("number", ""), text))
                    elements.append(Paragraph(text, style_n))
                    elements.append(Paragraph(answer, style_answer))

        return elements

    @staticmethod
    def get_localised_date(date_to_transform, timezone='Europe/London'):
        return arrow.get(date_to_transform).to(timezone).format("DD MMMM YYYY HH:mm:ss")


def parser(description=__doc__):
    rv = argparse.ArgumentParser(
        description,
    )
    rv.add_argument(
        "--survey", required=True,
        help="Set a path to the survey JSON file.")
    return rv


def main(args):
    fp = os.path.expanduser(os.path.abspath(args.survey))
    with open(fp, "r") as fobj:
        survey = json.load(fobj)

    data = json.load(sys.stdin)
    tx = PDFTransformer(survey, data)
    output = tx.render()
    sys.stdout.write(output.decode("latin-1"))
    return 0


def run():
    p = parser()
    args = p.parse_args()
    rv = main(args)
    sys.exit(rv)


if __name__ == "__main__":
    run()



#!/usr/bin/env python
#   coding: UTF-8

import os
import unittest

from sdx.common.config import aggregate_options
from sdx.common.config import check_safe_value
from sdx.common.config import config_parser


class AggregateOptionTests(unittest.TestCase):

    def test_settings_from_config(self):
        cfg = config_parser()
        cfg["sdx.collect"] = {"secret": "x" * 44}
        rv = aggregate_options(cfg, name="sdx.collect", keys=["secret"])
        self.assertEqual({"secret": "x" * 44}, rv)

    @unittest.skipIf(
        "SDX_COLLECT_SECRET" in os.environ,
        "variables match live environment"
    )
    def test_no_settings_only_env(self):
        try:
            os.environ["SDX_COLLECT_SECRET"] = "y" * 44
            self.assertTrue(os.getenv("SDX_COLLECT_SECRET"))
            cfg = config_parser()
            rv = aggregate_options(cfg, name="sdx.collect", keys=["secret"])
            self.assertEqual({"secret": "y" * 44}, rv)
        finally:
            del os.environ["SDX_COLLECT_SECRET"]

    @unittest.skipIf(
        "SDX_COLLECT_SECRET" in os.environ,
        "variables match live environment"
    )
    def test_env_overrides_settings(self):
        try:
            os.environ["SDX_COLLECT_SECRET"] = "y" * 44
            self.assertTrue(os.getenv("SDX_COLLECT_SECRET"))
            cfg = config_parser()
            cfg["sdx.collect"] = {"secret": "x" * 44}
            rv = aggregate_options(cfg, name="sdx.collect", keys=["secret"])
            self.assertEqual({"secret": "y" * 44}, rv)
        finally:
            del os.environ["SDX_COLLECT_SECRET"]

    def test_no_settings(self):
        cfg = config_parser()
        rv = aggregate_options(cfg, name="sdx.collect", keys=["secret"])
        self.assertEqual({}, rv)


class CheckSafeValueTests(unittest.TestCase):

    secretBytes = b"SKbH0bRYIa9irkLYxPkAi68JSt0_M1x3lci4nCIK7Ec="
    secretBytesWithDollar = b"SKbH0bRYIa9irkLYxPkAi68J$t0_M1x3lci4nCIK7Ec="
    secretBytesWithHash = b"SKb#0bRYIa9irkLYxPkAi68JSt0_M1x3lci4nCIK7Ec="
    secretString = "xr37-bYBRm1HYsJXuoq1X-gkt3WUxqgiSALpzuvtlSc="
    secretStringWithDollar = "xr37-bYBRm1HYsJXuoq1X-gkt3WUxqgiSALpzuvtl$c="
    secretStringWithHash = "xr37-bYBRm1#YsJXuoq1X-gkt3WUxqgiSALpzuvtlSc="

    def test_bytes(self):
        self.assertFalse(check_safe_value(CheckSafeValueTests.secretBytes))

    def test_bytes_with_dollar(self):
        self.assertIn(b"$", CheckSafeValueTests.secretBytesWithDollar)
        self.assertFalse(
            check_safe_value(CheckSafeValueTests.secretBytesWithDollar)
        )

    def test_bytes_with_hash(self):
        self.assertIn(b"#", CheckSafeValueTests.secretBytesWithHash)
        self.assertFalse(
            check_safe_value(CheckSafeValueTests.secretBytesWithHash)
        )

    def test_none(self):
        self.assertFalse(check_safe_value(None))

    def test_string(self):
        self.assertTrue(check_safe_value(CheckSafeValueTests.secretString))

    def test_string_with_dollar(self):
        self.assertIn("$", CheckSafeValueTests.secretStringWithDollar)
        self.assertFalse(
            check_safe_value(CheckSafeValueTests.secretStringWithDollar)
        )

    def test_string_with_hash(self):
        self.assertIn("#", CheckSafeValueTests.secretStringWithHash)
        self.assertFalse(
            check_safe_value(CheckSafeValueTests.secretStringWithHash)
        )



import logging
import os


def logger_initial_config(service_name=None,
                          log_level=None,
                          logger_format=None,
                          logger_date_format=None):
    '''Set initial logging configurations.

    :param service_name: Name of the service
    :type logger: String

    :param log_level: A string or integer corresponding to a Python logging level
    :type log_level: String

    :param logger_format: A string defining the format of the logs
    :type log_level: String

    :param logger_date_format: A string defining the format of the date/time in the logs
    :type log_level: String

    :rtype: None

    '''
    if not log_level:
        log_level = os.getenv('LOGGING_LEVEL', 'DEBUG')
    if not logger_format:
        logger_format = (
            "%(asctime)s.%(msecs)06dZ|"
            "%(levelname)s: {}: %(message)s"
        ).format(service_name)
    if not logger_date_format:
        logger_date_format = os.getenv('LOGGING_DATE_FORMAT', "%Y-%m-%dT%H:%M:%S")

    logging.basicConfig(level=log_level,
                        format=logger_format,
                        datefmt=logger_date_format)



import logging
import os


def set_level(logger=None, log_level=None):
    '''Set logging levels using logger names.

    :param logger: Name of the logger
    :type logger: String

    :param log_level: A string or integer corresponding to a Python logging level
    :type log_level: String

    :rtype: None

    '''
    log_level = logging.getLevelName(os.getenv('VERBOSITY', 'WARNING'))
    logging.getLogger(logger).setLevel(log_level)



from collections import namedtuple
import datetime
import json
import logging

import pkg_resources


__doc__ = """
The survey module provides functions to access survey reply data.
These will be of use in :ref:`processors` and :ref:`transformers`.

"""


class Survey:
    """Provide operations and accessors to survey data."""

    #: A named tuple type to capture ids and discriminators from a survey response.
    Identifiers = namedtuple("Identifiers", [
        "batch_nr", "seq_nr", "ts", "tx_id", "survey_id", "inst_id",
        "user_ts", "user_id", "ru_ref", "ru_check", "period"
    ])

    @staticmethod
    def load_survey(ids, package, pattern):
        """Retrieve the survey definition by id.

        This function takes metadata from a survey reply, finds the JSON definition of
        that survey, and loads it as a Python object.

        :param ids: Survey response ids.
        :type ids: :py:class:`sdx.common.survey.Survey.Identifiers`
        :param str package: The name of the Python package where the survey is to be found,
                            eg: `"sdx.common"`. Within standalone apps which do not declare
                            themselves as Python packages, pass in the variable `__name__`.
        :param str pattern: A query for the survey definition. This will be
                            a file path relative to the package location which uniquely
                            identifies the survey definition file. It accepts keyword
                            formatting arguments for any of the attributes of
                            :py:class:`sdx.common.survey.Survey.Identifiers`.

                            For example: `"surveys/{survey_id}.{inst_id}.json"`.
        :rtype: dict

        """
        try:
            content = pkg_resources.resource_string(
                package, pattern.format(**ids._asdict())
            )
        except FileNotFoundError:
            return None
        else:
            return json.loads(content.decode("utf-8"))

    @staticmethod
    def bind_logger(log, ids):
        """Bind a structured logger with survey response metadata.

        :param log: The logger object to be bound.
        :param ids: The survey response ids to bind to the logger.
        :type ids: :py:class:`sdx.common.survey.Survey.Identifiers`

        """
        return log.bind(
            ru_ref=ids.ru_ref,
            tx_id=ids.tx_id,
            user_id=ids.user_id,
        )

    @staticmethod
    def parse_timestamp(text):
        """Parse a text field for a date or timestamp.

        Date and time formats vary across surveys.
        This method reads those formats.

        :param str text: The date or timestamp value.
        :rtype: Python date or datetime.

        """

        cls = datetime.datetime

        if text.endswith("Z"):
            return cls.strptime(text, "%Y-%m-%dT%H:%M:%SZ").replace(
                tzinfo=datetime.timezone.utc
            )

        try:
            return cls.strptime(text, "%Y-%m-%dT%H:%M:%S.%f%z")
        except ValueError:
            pass

        try:
            return cls.strptime(text.partition(".")[0], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            pass

        try:
            return cls.strptime(text, "%Y-%m-%d").date()
        except ValueError:
            pass

        try:
            return cls.strptime(text, "%d/%m/%Y").date()
        except ValueError:
            pass

        if len(text) != 6:
            return None

        try:
            return cls.strptime(text + "01", "%Y%m%d").date()
        except ValueError:
            return None

    @staticmethod
    def identifiers(data, batch_nr=0, seq_nr=0, log=None):
        """Parse common metadata from the survey.

        Return a named tuple which code can use to access the various ids and discriminators.

        :param dict data:   A survey reply.
        :param int batch_nr: A batch number for the reply.
        :param int seq_nr: An image sequence number for the reply.

        """
        log = log or logging.getLogger(__name__)
        ru_ref = data.get("metadata", {}).get("ru_ref", "")
        ts = datetime.datetime.now(datetime.timezone.utc)
        rv = Survey.Identifiers(
            batch_nr, seq_nr, ts,
            data.get("tx_id"),
            data.get("survey_id"),
            data.get("collection", {}).get("instrument_id"),
            Survey.parse_timestamp(data.get("submitted_at", ts.isoformat())),
            data.get("metadata", {}).get("user_id"),
            ''.join(i for i in ru_ref if i.isdigit()),
            ru_ref[-1] if ru_ref and ru_ref[-1].isalpha() else "",
            data.get("collection", {}).get("period")
        )
        if any(i is None for i in rv):
            log.warning("Missing an id from {0}".format(rv))
            return None
        else:
            return rv



from collections import OrderedDict
import datetime
import json
import unittest

import pkg_resources

from sdx.common.formats.cs_formatter import CSFormatter
from sdx.common.survey import Survey


class BatchFileTests(unittest.TestCase):

    def test_pck_batch_header(self):
        batch_nr = 3866
        batch_date = datetime.date(2009, 12, 29)
        rv = CSFormatter.pck_batch_header(batch_nr, batch_date)
        self.assertEqual("FBFV00386629/12/09", rv)

    def test_pck_form_header(self):
        form_id = 5
        ru_ref = 49900001225
        check = "C"
        period = "200911"
        rv = CSFormatter.pck_form_header(form_id, ru_ref, check, period)
        self.assertEqual("0005:49900001225C:200911", rv)

    def test_pck_lines(self):
        batch_nr = 3866
        batch_date = datetime.date(2009, 12, 29)
        survey_id = "134"
        inst_id = "0005"
        ru_ref = 49900001225
        check = "C"
        period = "200911"
        data = OrderedDict([
            ("0001", 2),
            ("0140", 124),
            ("0151", 217222)
        ])
        self.assertTrue(isinstance(val, int) for val in data.values())
        rv = CSFormatter.pck_lines(
            data, batch_nr, batch_date, survey_id, inst_id, ru_ref, check, period
        )
        self.assertEqual([
            "FV          ",
            "0005:49900001225C:200911",
            "0001 00000000002",
            "0140 00000000124",
            "0151 00000217222",
        ], rv)

    def test_idbr_receipt(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        reply["tx_id"] = "27923934-62de-475c-bc01-433c09fd38b8"
        ids = Survey.identifiers(reply, batch_nr=3866)
        rv = CSFormatter.idbr_receipt(**ids._asdict())
        self.assertEqual("12346789012:A:134:201605", rv)

    def test_identifiers(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        reply["tx_id"] = "27923934-62de-475c-bc01-433c09fd38b8"
        reply["collection"]["period"] = "200911"
        ids = Survey.identifiers(reply)
        self.assertIsInstance(ids, Survey.Identifiers)
        self.assertEqual(0, ids.batch_nr)
        self.assertEqual(0, ids.seq_nr)
        self.assertEqual(reply["tx_id"], ids.tx_id)
        self.assertEqual(datetime.date.today(), ids.ts.date())
        self.assertEqual("134", ids.survey_id)
        self.assertEqual("K5O86M2NU1", ids.user_id)
        self.assertEqual("12346789012", ids.ru_ref)
        self.assertEqual("A", ids.ru_check)
        self.assertEqual("200911", ids.period)

    def test_pck_from_transformed_data(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        reply["tx_id"] = "27923934-62de-475c-bc01-433c09fd38b8"
        reply["survey_id"] = "134"
        reply["collection"]["period"] = "200911"
        reply["metadata"]["ru_ref"] = "49900001225C"
        reply["data"] = OrderedDict([
            ("0001", 2),
            ("0140", 124),
            ("0151", 217222)
        ])
        ids = Survey.identifiers(reply, batch_nr=3866)
        rv = CSFormatter.pck_lines(reply["data"], **ids._asdict())
        self.assertEqual([
            "FV          ",
            "0005:49900001225C:200911",
            "0001 00000000002",
            "0140 00000000124",
            "0151 00000217222",
        ], rv)



import unittest

from sdx.common.processor import Processor


class ProcessorTests(unittest.TestCase):

    def test_processor_unsigned(self):
        proc = Processor.unsigned_integer

        # Supply int default for range checking
        self.assertEqual(0, proc("q", {"q": -1}, 0))
        self.assertEqual(0, proc("q", {"q": 0}, 0))
        self.assertEqual(1, proc("q", {"q": 1}, 0))
        self.assertEqual(100, proc("q", {"q": 1E2}, 0))
        self.assertEqual(1000000000, proc("q", {"q": 1E9}, 0))

        # Supply bool default for range checking and type coercion
        self.assertIs(False, proc("q", {"q": -1}, False))
        self.assertIs(False, proc("q", {"q": 0}, False))
        self.assertIs(True, proc("q", {"q": 1}, False))
        self.assertIs(True, proc("q", {"q": 1E2}, False))
        self.assertIs(True, proc("q", {"q": 1E9}, False))
        self.assertIs(False, proc("q", {"q": 0}, False))

    def test_processor_percentage(self):
        proc = Processor.percentage

        # Supply int default for range checking
        self.assertEqual(0, proc("q", {"q": -1}, 0))
        self.assertEqual(0, proc("q", {"q": 0}, 0))
        self.assertEqual(100, proc("q", {"q": 100}, 0))
        self.assertEqual(0, proc("q", {"q": 0}, 0))

        # Supply bool default for range checking and type coercion
        self.assertIs(False, proc("q", {"q": -1}, False))
        self.assertIs(False, proc("q", {"q": 0}, False))
        self.assertIs(True, proc("q", {"q": 100}, False))
        self.assertIs(False, proc("q", {"q": 0}, False))



#!/usr/bin/env python
#   coding: UTF-8

import glob
from io import BytesIO
import os.path
import shutil
import subprocess
import zipfile

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.exceptions import MaxRetryError
from requests.packages.urllib3.util.retry import Retry

from sdx.common.survey import Survey

__doc__ = """
The SDX Image Transformer:

"""


class ImageTransformer(object):

    session = requests.Session()

    @staticmethod
    def extract_pdf_images(path, f_name):
        """Extract pages from a PDF document.

        :param str path: The location of the working directory.
        :param str f_name: The file name of the PDF document.
        :return: A sorted sequence of image file names.

        This method delegates to the *pdftoppm* utility.

        """
        rootName, _ = os.path.splitext(f_name)
        subprocess.call(
            ["pdftoppm", "-jpeg", f_name, rootName],
            cwd=path
        )
        return sorted(glob.glob("%s/%s-*.jpg" % (path, rootName)))

    def __init__(self, logger, settings, survey, response_data, sequence_no=1000):
        self.logger = logger
        self.settings = settings
        self.survey = survey
        self.response = response_data
        self.sequence_no = sequence_no

        adapter = self.session.adapters["http://"]
        if adapter.max_retries.total != 5:
            retries = Retry(total=5, backoff_factor=0.1)
            self.session.mount("http://", HTTPAdapter(max_retries=retries))
            self.session.mount("https://", HTTPAdapter(max_retries=retries))

    def get_image_sequence_numbers(self):
        sequence_numbers = []
        for image in self.images:
            sequence_number = self.get_image_sequence_no()
            sequence_numbers.append(sequence_number)

        self.logger.debug('Sequence numbers generated', sequence_numbers=sequence_numbers)
        return sequence_numbers

    def create_image_sequence(self, path, nmbr_seq=None):
        """Renumber the image sequence extracted from pdf

        :param str path: The location of the working directory.
        :param nmbr_seq: A sequence or generator of integers.
        :type nmbr_seq: list or generator.
        :return: A generator of file paths.

        """
        locn, baseName = os.path.split(path)
        self.images = ImageTransformer.extract_pdf_images(locn, baseName)
        nmbr_seq = nmbr_seq or self.get_image_sequence_numbers()
        for imageFile, n in zip(self.images, nmbr_seq):
            name = "S%09d.JPG" % n
            fp = os.path.join(locn, name)
            os.rename(imageFile, fp)
            yield fp

    def image_path(self, img):
        return "\\".join([
            self.settings.FTP_HOST.rstrip("\\"),
            self.settings.SDX_FTP_IMAGE_PATH.strip("\\"),
            "Images",
            os.path.basename(img)
        ])

    def index_lines(self, ids, images):
        return (
            ",".join([
                ids.ts.strftime("%d/%m/%Y %H:%M:%S"),
                self.image_path(img),
                ids.ts.strftime("%Y%m%d"),
                os.path.splitext(os.path.basename(img))[0],
                ids.survey_id,
                ids.inst_id,
                ids.ru_ref,
                Survey.parse_timestamp(ids.period).strftime("%Y%m"),
                "{0:03}".format(n + 1) if n else "001,0"
            ])
            for n, img in enumerate(images)
        )

    def create_image_index(self, images):
        '''
        Takes a list of images and creates a index csv from them
        '''
        if not images:
            return None

        ids = Survey.identifiers(self.response)

        for i in images:
            self.logger.info("Adding image to index", file=(self.image_path(i)))

        self.index_file = "EDC_{0}_{1}_{2:04}.csv".format(
            ids.survey_id,
            ids.user_ts.strftime("%Y%m%d"),
            self.sequence_no
        )

        locn = os.path.dirname(images[0])
        path = os.path.join(locn, self.index_file)
        with open(path, "w") as fh:
            fh.write("\n".join(self.index_lines(ids, images)))
        return path

    def create_zip(self, images, index):
        '''
        Create a zip from a renumbered sequence
        '''
        in_memory_zip = BytesIO()

        with zipfile.ZipFile(in_memory_zip, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for fp in images:
                f_name = os.path.basename(fp)
                try:
                    zipf.write(fp, arcname=f_name)
                except Exception as e:
                    self.logger.error(e)

            if index:
                f_name = os.path.basename(index)
                zipf.write(index, arcname=f_name)

        # Return to beginning of file
        in_memory_zip.seek(0)

        return in_memory_zip

    def cleanup(self, locn):
        '''
        Remove all temporary files, by removing top level dir
        '''
        shutil.rmtree(locn)

    def response_ok(self, res):

        if res.status_code == 200:
            self.logger.info(
                "Returned from sdx-sequence",
                request_url=res.url,
                status=res.status_code
            )
            return True
        else:
            self.logger.error(
                "Returned from sdx-sequence",
                request_url=res.url,
                status=res.status_code
            )
            return False

    def remote_call(self, request_url, json=None):
        try:
            self.logger.info("Calling sdx-sequence", request_url=request_url)

            r = None

            if json:
                r = self.session.post(request_url, json=json)
            else:
                r = self.session.get(request_url)

            return r
        except MaxRetryError:
            self.logger.error("Max retries exceeded (5)", request_url=request_url)

    def get_image_sequence_no(self):
        sequence_url = self.settings.SDX_SEQUENCE_URL + "/image-sequence"

        r = self.remote_call(sequence_url)

        if not self.response_ok(r):
            return False

        result = r.json()
        return result['sequence_no']



from collections import namedtuple
from collections import OrderedDict
import datetime
from functools import partial
import itertools
import json
import os.path
import re
import unittest
import zipfile

import pkg_resources

from sdx.common.formats.cs_formatter import CSFormatter
from sdx.common.processor import Processor
from sdx.common.survey import Survey
from sdx.common.transformer import Transformer
from sdx.common.transforms.ImageTransformer import ImageTransformer


class TestTransformer(Transformer):
    """An example of how to build a transformer.

    Taken from the MWSS survey.

    """

    defn = [
        (40, 0, partial(Processor.aggregate, weights=[("40f", 1)])),
        (50, 0, partial(Processor.aggregate, weights=[("50f", 0.5)])),
        (60, 0, partial(Processor.aggregate, weights=[("60f", 0.5)])),
        (70, 0, partial(Processor.aggregate, weights=[("70f", 0.5)])),
        (80, 0, partial(Processor.aggregate, weights=[("80f", 0.5)])),
        (90, False, partial(
            Processor.evaluate,
            group=[
                "90w", "91w", "92w", "93w", "94w", "95w", "96w", "97w",
                "90f", "91f", "92f", "93f", "94f", "95f", "96f", "97f",
            ],
            convert=re.compile("^((?!No).)+$").search, op=lambda x, y: x or y)),
        (100, False, partial(Processor.mean, group=["100f"])),
        (110, [], partial(Processor.events, group=["110f"])),
        (120, False, partial(Processor.mean, group=["120f"])),
        (range(130, 133, 1), False, Processor.survey_string),
        (140, 0, partial(
            Processor.aggregate,
            weights=[
                ("140m", 1), ("140w4", 1), ("140w5", 1)
            ])),
        (range(151, 154, 1), 0, Processor.unsigned_integer),
        (range(171, 174, 1), 0, Processor.unsigned_integer),
        (range(181, 184, 1), 0, Processor.unsigned_integer),
        (190, False, partial(
            Processor.evaluate,
            group=[
                "190w4", "191w4", "192w4", "193w4", "194w4", "195w4", "196w4", "197w4",
                "190m", "191m", "192m", "193m", "194m", "195m", "196m", "197m",
                "190w5", "191w5", "192w5", "193w5", "194w5", "195w5", "196w5", "197w5",
            ],
            convert=re.compile("^((?!No).)+$").search, op=lambda x, y: x or y)),
        (200, False, partial(Processor.mean, group=["200w4", "200w5"])),
        (210, [], partial(Processor.events, group=["210w4", "210w5"])),
        (220, False, partial(Processor.mean, group=["220w4", "220w5"])),
        (300, False, partial(
            Processor.evaluate,
            group=[
                "300w", "300f", "300m", "300w4", "300w5",
            ],
            convert=str, op=lambda x, y: x + "\n" + y)),
    ]

    package = "sdx.common.test"
    pattern = "data/{survey_id}.{inst_id}.json"


class LogicTests(unittest.TestCase):

    def test_weekly_increase(self):
        """
        Increase in weekly pay (100).

        """
        dflt, fn = TestTransformer.ops()["100"]
        rv = fn("100", {"100": "6.0"}, 0)
        self.assertEqual(6, rv)

    def test_aggregate_weekly_paid_employees(self):
        """
        QIds 40, 40f are added to give a value for weekly paid employees (40).

        """
        dflt, fn = TestTransformer.ops()["40"]
        rv = fn("40", {"40": "125000"}, 0)
        self.assertEqual(125000, rv)
        rv = fn("40", {"40": "125000", "40f": "25000"}, 0)
        self.assertEqual(150000, rv)

    def test_aggregate_fortnightly_gross_pay(self):
        """
        Fortnightly gross pay (50f) is divided by 2 and added to
        the value for qid 50.

        """
        dflt, fn = TestTransformer.ops()["50"]
        rv = fn("50", {"50f": "1600"}, 0)
        self.assertEqual(800, rv)
        rv = fn("50", {"50": "19200", "50f": "1600"}, 0)
        self.assertEqual(20000, rv)

    def test_aggregate_fortnightly_bonuses(self):
        """
        Fortnightly holiday pay (60f), arrears of pay (70f) and bonuses (80f)
        are divided by 2 and added to qids 60, 70, 80 respectively.

        """
        dflt, fn = TestTransformer.ops()["60"]
        rv = fn("60", {"60f": "360"}, 0)
        self.assertEqual(180, rv)
        rv = fn("60", {"60": "4600", "60f": "360"}, 0)
        self.assertEqual(4780, rv)

        dflt, fn = TestTransformer.ops()["70"]
        rv = fn("70", {"70f": "1280"}, 0)
        self.assertEqual(640, rv)
        rv = fn("70", {"70": "7360", "70f": "1280"}, 0)
        self.assertEqual(8000, rv)

        dflt, fn = TestTransformer.ops()["80"]
        rv = fn("80", {"80f": "5000"}, 0)
        self.assertEqual(2500, rv)
        rv = fn("80", {"80": "15000", "80f": "5000"}, 0)
        self.assertEqual(17500, rv)

    def test_aggregate_fortnightly_increase(self):
        """
        Increase in Fortnightly pay (100f); aggregated with weekly increase (100).

        """
        dflt, fn = TestTransformer.ops()["100"]
        rv = fn("100", {"100f": "6.0"}, 0)
        self.assertEqual(6, rv)
        rv = fn("100", {"100": "7.0", "100f": "6.0"}, 0)
        self.assertEqual(6, rv)  # Integer default
        rv = fn("100", {"100": "7.0", "100f": "6.0"}, 0.0)
        self.assertEqual(6.5, rv)  # Float default

    def test_aggregate_fortnightly_increase_date(self):
        """
        Date of increase in Fortnightly pay (110f); aggregated with weekly (110).

        """
        dflt, fn = TestTransformer.ops()["110"]
        rv = fn(
            "110", {"110": "2017-01-09", "110f": "2017-01-11"}, datetime.date.today(),
        )
        self.assertEqual(9, rv[0].day)
        self.assertEqual(11, rv[1].day)

    def test_aggregate_fortnightly_increase_employees(self):
        """
        Employees with increase in Fortnightly pay (120f);
        aggregated with weekly increase (120).

        """
        dflt, fn = TestTransformer.ops()["120"]
        rv = fn("120", {"120f": "60"}, 0)
        self.assertEqual(60, rv)
        rv = fn("120", {"120": "40", "120f": "41"}, 0)
        self.assertEqual(40, rv)  # Integer default
        rv = fn("120", {"120": "40", "120f": "41"}, 0.0)
        self.assertEqual(40.5, rv)  # Float default

    def test_aggregate_fortnightly_changes(self):
        """
        QIds 90f - 97f used for fortnightly changes questions; all aggregated as 90.

        """
        dflt, fn = TestTransformer.ops()["90"]
        for qid in ("90f", "91f", "92f", "93f", "94f", "95f", "96f", "97f"):
            with self.subTest(qid=qid):
                rv = fn("90", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("90", {qid: "No"}, True)
                self.assertFalse(rv)
                rv = fn("90", {qid: "Yes"}, False)
                self.assertTrue(rv)

    def test_aggregate_weekly_changes(self):
        """
        QIds 90w - 97w used for weekly changes questions; all aggregated as 90.

        """
        dflt, fn = TestTransformer.ops()["90"]
        for qid in ("90w", "91w", "92w", "93w", "94w", "95w", "96w", "97w"):
            with self.subTest(qid=qid):
                rv = fn("90", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("90", {qid: "No"}, True)
                self.assertFalse(rv)
                rv = fn("90", {qid: "Yes"}, False)
                self.assertTrue(rv)

    def test_radio_button_logic(self):
        """
        QIds 92w, 94w, 92f, 94f, 192m, 194m, 192w4, 194w4, 192w5, 194w5
        have answers other than Yes/No.

        """
        dflt, fn = TestTransformer.ops()["90"]
        for qid in ("92w", "94w", "92f", "94f"):
            with self.subTest(qid=qid):
                rv = fn("90", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("90", {qid: "No significant change"}, True)
                self.assertFalse(rv)
                rv = fn("90", {qid: "Any other string"}, False)
                self.assertTrue(rv)

        dflt, fn = TestTransformer.ops()["190"]
        for qid in ("192m", "194m", "192w4", "194w4", "192w5", "194w5"):
            with self.subTest(qid=qid):
                rv = fn("190", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "No significant change"}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "Any other string"}, False)
                self.assertTrue(rv)

    def test_aggregate_fourweekly_changes(self):
        """
        QIds 190w4 - 197w4 used for fourweekly changes questions; all aggregated as 190.

        """
        dflt, fn = TestTransformer.ops()["190"]
        for qid in ("190w4", "191w4", "192w4", "193w4", "194w4", "195w4", "196w4", "197w4"):
            with self.subTest(qid=qid):
                rv = fn("190", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "No"}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "Yes"}, False)
                self.assertTrue(rv)

    def test_aggregate_fourweekly_increase(self):
        """
        Increase in fourweekly pay (200w4); aggregated with monthly increase (200).

        """
        dflt, fn = TestTransformer.ops()["200"]
        rv = fn("200", {"200w4": "6.0"}, 0)
        self.assertEqual(6, rv)
        rv = fn("200", {"200": "7.0", "200w4": "6.0"}, 0)
        self.assertEqual(6, rv)  # Integer default
        rv = fn("200", {"200": "7.0", "200w4": "6.0"}, 0.0)
        self.assertEqual(6.5, rv)  # Float default

    def test_aggregate_fourweekly_increase_date(self):
        """
        Date of increase in fourweekly pay (210w4); aggregated with monthly (210).

        """
        dflt, fn = TestTransformer.ops()["210"]
        rv = fn(
            "210", {"210": "2017-01-09", "210w4": "2017-01-11"}, datetime.date.today(),
        )
        self.assertEqual(9, rv[0].day)
        self.assertEqual(11, rv[1].day)

    def test_aggregate_fourweekly_increase_employees(self):
        """
        Employees with increase in fourweekly pay (220w4);
        aggregated with monthly increase (220).

        """
        dflt, fn = TestTransformer.ops()["220"]
        rv = fn("220", {"220w4": "60"}, 0)
        self.assertEqual(60, rv)
        rv = fn("220", {"220": "40", "220w4": "41"}, 0)
        self.assertEqual(40, rv)  # Integer default
        rv = fn("220", {"220": "40", "220w4": "41"}, 0.0)
        self.assertEqual(40.5, rv)  # Float default

    def test_aggregate_monthly_changes(self):
        """
        QIds 190m - 197m used for monthly changes questions; all aggregated as 190.

        """
        dflt, fn = TestTransformer.ops()["190"]
        for qid in ("190m", "191m", "192m", "193m", "194m", "195m", "196m", "197m"):
            with self.subTest(qid=qid):
                rv = fn("190", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "No"}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "Yes"}, False)
                self.assertTrue(rv)

    def test_aggregate_weekly_comments(self):
        """
        QIds 300w, 300f, 300m, 300w4 & 300w5; all aggregated as 300.

        """
        dflt, fn = TestTransformer.ops()["300"]
        for qid in ("300w", "300f", "300m", "300w4", "300w5"):
            with self.subTest(qid=qid):
                rv = fn("300", {qid: "Single comment"}, "")
                self.assertEqual("Single comment", rv)
                rv = fn("300", {"300": "First comment", qid: "Second comment"}, "")
                self.assertEqual(["First comment", "Second comment"], rv.splitlines())

    def test_aggregate_monthly_paid_employees(self):
        """
        QIds 140m, 140w4, 140w5 are added to give a value for monthly paid employees (140).

        """
        dflt, fn = TestTransformer.ops()["140"]
        rv = fn("140", {"140w4": "125000"}, 0)
        self.assertEqual(125000, rv)
        for qid in ("140m", "140w4", "140w5"):
            rv = fn("140", {"140": "125000", qid: "25000"}, 0)
            self.assertEqual(150000, rv)

    def test_aggregate_fiveweekly_changes(self):
        """
        QIds 190w5 - 197w5 used for fiveweekly changes questions; all aggregated as 190.

        """
        dflt, fn = TestTransformer.ops()["190"]
        for qid in ("190w5", "191w5", "192w5", "193w5", "194w5", "195w5", "196w5", "197w5"):
            with self.subTest(qid=qid):
                rv = fn("190", {qid: ""}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "No"}, True)
                self.assertFalse(rv)
                rv = fn("190", {qid: "Yes"}, False)
                self.assertTrue(rv)

    def test_aggregate_fiveweekly_increase(self):
        """
        Increase in fiveweekly pay (200w5); aggregated with monthly increase (200).

        """
        dflt, fn = TestTransformer.ops()["200"]
        rv = fn("200", {"200w5": "6.0"}, 0)
        self.assertEqual(6, rv)
        rv = fn("200", {"200": "7.0", "200w5": "6.0"}, 0)
        self.assertEqual(6, rv)  # Integer default
        rv = fn("200", {"200": "7.0", "200w5": "6.0"}, 0.0)
        self.assertEqual(6.5, rv)  # Float default

    def test_aggregate_fiveweekly_increase_date(self):
        """
        Date of increase in fiveweekly pay (210w5); aggregated with monthly (210).

        """
        dflt, fn = TestTransformer.ops()["210"]
        rv = fn(
            "210", {"210": "2017-01-09", "210w5": "2017-01-11"}, datetime.date.today(),
        )
        self.assertEqual(9, rv[0].day)
        self.assertEqual(11, rv[1].day)

    def test_aggregate_fiveweekly_increase_employees(self):
        """
        Employees with increase in fiveweekly pay (220w5);
        aggregated with monthly increase (220).

        """
        dflt, fn = TestTransformer.ops()["220"]
        rv = fn("220", {"220w5": "60"}, 0)
        self.assertEqual(60, rv)
        rv = fn("220", {"220": "40", "220w5": "41"}, 0)
        self.assertEqual(40, rv)  # Integer default
        rv = fn("220", {"220": "40", "220w5": "41"}, 0.0)
        self.assertEqual(40.5, rv)  # Float default


class TransformTests(unittest.TestCase):

    def test_unsigned(self):
        rv = TestTransformer.transform({"40": "33"})
        self.assertEqual(33, rv["40"])
        item = CSFormatter.pck_item("40", rv["40"])
        self.assertEqual(item, "0040 00000000033")

    def test_currency(self):
        rv = TestTransformer.transform({"50": "36852"})
        self.assertEqual(36852, rv["50"])
        item = CSFormatter.pck_item("50", rv["50"])
        self.assertEqual(item, "0050 00000036852")

    def test_digits_to_onetwo(self):
        digits_ingested_as_bools = [100, 120, 200, 220]
        for qNr in digits_ingested_as_bools:
            qid = str(qNr)
            with self.subTest(qNr=qNr, qid=qid):
                rv = TestTransformer.transform({qid: "64"})
                self.assertIs(True, rv[qid])
                self.assertEqual(1, CSFormatter.pck_value(qid, rv[qid]))
                rv = TestTransformer.transform({qid: ""})
                self.assertEqual(2, CSFormatter.pck_value(qid, rv[qid]))

    def test_dates_to_onetwo(self):
        dates_ingested_as_bools = [110, 210]
        for qNr in dates_ingested_as_bools:
            qid = str(qNr)
            with self.subTest(qNr=qNr, qid=qid):
                rv = TestTransformer.transform({qid: "23/4/2017"})
                self.assertEqual([datetime.date(2017, 4, 23)], rv[qid])
                self.assertEqual(1, CSFormatter.pck_value(qid, rv[qid]))
                rv = TestTransformer.transform({qid: ""})
                self.assertEqual([], rv[qid])
                self.assertEqual(2, CSFormatter.pck_value(qid, rv[qid]))


class BatchFileTests(unittest.TestCase):

    def test_pck_batch_header(self):
        batch_nr = 3866
        batch_date = datetime.date(2009, 12, 29)
        rv = CSFormatter.pck_batch_header(batch_nr, batch_date)
        self.assertEqual("FBFV00386629/12/09", rv)

    def test_pck_form_header(self):
        form_id = 5
        ru_ref = 49900001225
        check = "C"
        period = "200911"
        rv = CSFormatter.pck_form_header(form_id, ru_ref, check, period)
        self.assertEqual("0005:49900001225C:200911", rv)

    def test_load_survey(self):
        ids = Survey.identifiers({
            "survey_id": "134",
            "tx_id": "27923934-62de-475c-bc01-433c09fd38b8",
            "collection": {
                "instrument_id": "0005",
                "period": "201704"
            },
            "metadata": {
                "user_id": "123456789",
                "ru_ref": "12345678901A"
            }
        })
        rv = Survey.load_survey(ids, TestTransformer.package, TestTransformer.pattern)
        self.assertIsNotNone(rv)

    def test_load_survey_miss(self):
        ids = Survey.identifiers({
            "survey_id": "127",
            "tx_id": "27923934-62de-475c-bc01-433c09fd38b8",
            "collection": {
                "instrument_id": "0001",
                "period": "201704"
            },
            "metadata": {
                "user_id": "123456789",
                "ru_ref": "12345678901A"
            }
        })
        rv = Survey.load_survey(ids, TestTransformer.package, TestTransformer.pattern)
        self.assertIsNone(rv)

    def test_image_path(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        tfr = TestTransformer(reply)
        survey = Survey.load_survey(tfr.ids, tfr.package, tfr.pattern)
        settings = PackingTests.Settings("\\NFS", "SDX")
        for settings in [
            PackingTests.Settings("\\\\NFS", "SDX"),
            PackingTests.Settings("\\\\NFS\\", "SDX"),
            PackingTests.Settings("\\\\NFS", "\\SDX"),
            PackingTests.Settings("\\\\NFS", "SDX\\"),
            PackingTests.Settings("\\\\NFS\\", "SDX\\"),
            PackingTests.Settings("\\\\NFS\\", "\\SDX\\"),
        ]:
            with self.subTest(settings=settings):
                img_tfr = ImageTransformer(
                    tfr.log, settings, survey, tfr.response, tfr.ids.seq_nr
                )
                self.assertEqual(
                    "\\\\NFS\\SDX\\Images\\IMG_01.JPG",
                    img_tfr.image_path("/work/IMG_01.JPG")
                )

    def test_pck_lines(self):
        batch_nr = 3866
        batch_date = datetime.date(2009, 12, 29)
        survey_id = "134"
        inst_id = "0005"
        ru_ref = 49900001225
        check = "C"
        period = "200911"
        data = OrderedDict([
            ("0001", 2),
            ("0140", 124),
            ("0151", 217222)
        ])
        self.assertTrue(isinstance(val, int) for val in data.values())
        rv = CSFormatter.pck_lines(
            data, batch_nr, batch_date, survey_id, inst_id, ru_ref, check, period
        )
        self.assertEqual([
            "FV          ",
            "0005:49900001225C:200911",
            "0001 00000000002",
            "0140 00000000124",
            "0151 00000217222",
        ], rv)

    def test_idbr_receipt(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        reply["tx_id"] = "27923934-62de-475c-bc01-433c09fd38b8"
        ids = Survey.identifiers(reply, batch_nr=3866)
        rv = CSFormatter.idbr_receipt(**ids._asdict())
        self.assertEqual("12346789012:A:134:201605", rv)

    def test_identifiers(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        reply["tx_id"] = "27923934-62de-475c-bc01-433c09fd38b8"
        reply["collection"]["period"] = "200911"
        ids = Survey.identifiers(reply)
        self.assertIsInstance(ids, Survey.Identifiers)
        self.assertEqual(0, ids.batch_nr)
        self.assertEqual(0, ids.seq_nr)
        self.assertEqual(reply["tx_id"], ids.tx_id)
        self.assertEqual(datetime.date.today(), ids.ts.date())
        self.assertEqual("134", ids.survey_id)
        self.assertEqual("K5O86M2NU1", ids.user_id)
        self.assertEqual("12346789012", ids.ru_ref)
        self.assertEqual("A", ids.ru_check)
        self.assertEqual("200911", ids.period)

    def test_pck_from_transformed_data(self):
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        reply["tx_id"] = "27923934-62de-475c-bc01-433c09fd38b8"
        reply["survey_id"] = "134"
        reply["collection"]["period"] = "200911"
        reply["metadata"]["ru_ref"] = "49900001225C"
        reply["data"] = OrderedDict([
            ("0001", 2),
            ("0140", 124),
            ("0151", 217222)
        ])
        ids = Survey.identifiers(reply, batch_nr=3866)
        rv = CSFormatter.pck_lines(reply["data"], **ids._asdict())
        self.assertEqual([
            "FV          ",
            "0005:49900001225C:200911",
            "0001 00000000002",
            "0140 00000000124",
            "0151 00000217222",
        ], rv)


class PackingTests(unittest.TestCase):

    Settings = namedtuple(
        "Settings",
        [
            "FTP_HOST",
            "SDX_FTP_IMAGE_PATH",
        ]
    )

    def test_mwss_pack(self):
        settings = PackingTests.Settings("\\NFS", "SDX")
        src = pkg_resources.resource_string("sdx.common.test", "data/eq-mwss.json")
        reply = json.loads(src.decode("utf-8"))
        tfr = TestTransformer(reply)
        self.assertEqual(
            "REC0103_0000.DAT",
            CSFormatter.idbr_name(
                **tfr.ids._asdict()
            )
        )
        zip_data = tfr.pack(settings=settings, img_seq=itertools.count(), tmp=None)

        with zipfile.ZipFile(zip_data) as zip_file:
            actual = zip_file.namelist()

        expected = [
            "EDC_QData/134_0000",
            "EDC_QReceipts/REC0103_0000.DAT",
            "EDC_QImages/Images/S000000000.JPG",
            "EDC_QImages/Images/S000000001.JPG",
            "EDC_QImages/Images/S000000002.JPG",
            "EDC_QImages/Images/S000000003.JPG",
            "EDC_QImages/Images/S000000004.JPG",
            "EDC_QImages/Images/S000000005.JPG",
            "EDC_QImages/Images/S000000006.JPG",
            "EDC_QImages/Images/S000000007.JPG",
            "EDC_QImages/Images/S000000008.JPG",
            "EDC_QImages/Images/S000000009.JPG",
            "EDC_QImages/Images/S000000010.JPG",
            "EDC_QImages/Index/EDC_134_20170301_0000.csv"
        ]
        self.assertEqual(expected, actual)

    def test_image_sequence_number(self):
        settings = PackingTests.Settings("\\NFS", "SDX")
        response = {
            "survey_id": "134",
            "tx_id": "27923934-62de-475c-bc01-433c09fd38b8",
            "collection": {
                "instrument_id": "0005",
                "period": "201704"
            },
            "metadata": {
                "user_id": "123456789",
                "ru_ref": "12345678901A"
            },
            "submitted_at": "2017-04-12T13:01:26Z",
            "data": {}
        }
        seq_nr = 12345
        tfr = TestTransformer(response, seq_nr=seq_nr)
        zf = zipfile.ZipFile(tfr.pack(settings=settings, img_seq=itertools.count(), tmp=None))
        fn = next(i for i in zf.namelist() if os.path.splitext(i)[1] == ".csv")
        bits = os.path.splitext(fn)[0].split("_")
        self.assertEqual(seq_nr, int(bits[-1]))



#!/usr/bin/env python
#  coding: UTF-8

import argparse
import logging
import os.path
import sys

DFLT_LOCN = os.path.expanduser("~")

__doc__ = """
Defines a common CLI for SDX tooling.

Operation via CLI requires a set of common options.


"""


def logging_format(service="sdx-common"):
    return "%(asctime)s|%(levelname)s: {0}: %(message)s".format(service)


def add_common_options(parser):
    parser.add_argument(
        "--version", action="store_true", default=False,
        help="Print the current version number")
    parser.add_argument(
        "-v", "--verbose", required=False,
        action="store_const", dest="log_level",
        const=logging.DEBUG, default=logging.INFO,
        help="Increase the verbosity of output")
    return parser


def add_transformer_options(parser):
    parser.add_argument(
        "--work", default=DFLT_LOCN,
        help="Set a path to the working directory.")
    parser.add_argument(
        "--img_nr", type=int, default=0,
        help="Set a starting number for the image sequence.")
    parser.add_argument(
        "--seq_nr", type=int, default=0,
        help="Set a sequence number for the data output.")
    parser.add_argument(
        "input",
        nargs="?", type=argparse.FileType("r"), default=sys.stdin,
        help="Specify survey data."
    )
    parser.add_argument(
        "output",
        nargs="?", type=argparse.FileType("wb"), default=sys.stdout.buffer,
        help="Specify output file."
    )
    return parser


def parser(description):
    return argparse.ArgumentParser(
        description,
    )


def transformer_cli(description=__doc__):
    return add_transformer_options(
        add_common_options(
            parser(description)
        )
    )



class CSFormatter:
    """Formatter for common software systems.

    Serialises standard data types to PCK format.
    Creates a receipt in IDBR format.

    """

    form_ids = {
        "134": "0004",
    }

    @staticmethod
    def pck_name(survey_id, seq_nr, **kwargs):
        """Generate the name of a PCK file."""
        return "{0}_{1:04}".format(survey_id, int(seq_nr))

    @staticmethod
    def pck_batch_header(batch_nr, ts):
        """Generate a batch header for a PCK file."""
        return "{0}{1:06}{2}".format("FBFV", batch_nr, ts.strftime("%d/%m/%y"))

    @staticmethod
    def pck_form_header(form_id, ru_ref, ru_check, period):
        """Generate a form header for PCK data."""
        form_id = "{0:04}".format(form_id) if isinstance(form_id, int) else form_id
        return "{0}:{1}{2}:{3}".format(form_id, ru_ref, ru_check, period)

    @staticmethod
    def pck_value(qid, val, survey_id=None):
        """Format a value as PCK data."""
        if isinstance(val, list):
            val = bool(val)

        if isinstance(val, bool):
            return 1 if val else 2
        elif isinstance(val, str):
            return 1 if val else 0
        else:
            return val

    @staticmethod
    def pck_item(q, a):
        """Return a PCK line item."""
        try:
            return "{0:04} {1:011}".format(int(q), CSFormatter.pck_value(q, a))
        except TypeError:
            return "{0} ???????????".format(q)

    @staticmethod
    def pck_lines(data, batch_nr, ts, survey_id, inst_id, ru_ref, ru_check, period, **kwargs):
        """Return a list of lines in a PCK file."""
        return [
            "FV" + " " * 10,
            CSFormatter.pck_form_header(inst_id, ru_ref, ru_check, period),
        ] + [
            CSFormatter.pck_item(q, a) for q, a in data.items()
        ]

    @staticmethod
    def write_pck(f_obj, data, **kwargs):
        """Write a PCK file."""
        output = CSFormatter.pck_lines(data, **kwargs)
        f_obj.write("\n".join(output))
        f_obj.write("\n")

    @staticmethod
    def idbr_name(user_ts, seq_nr, **kwargs):
        """Generate the name of an IDBR file."""
        return "REC{0}_{1:04}.DAT".format(user_ts.strftime("%d%m"), int(seq_nr))

    @staticmethod
    def idbr_receipt(survey_id, ru_ref, ru_check, period, **kwargs):
        """Format a receipt in IDBR format."""
        return "{ru_ref}:{ru_check}:{survey_id:03}:{period}".format(
            survey_id=int(survey_id), ru_ref=ru_ref, ru_check=ru_check, period=period
        )

    @staticmethod
    def write_idbr(f_obj, **kwargs):
        """Write an IDBR file."""
        output = CSFormatter.idbr_receipt(**kwargs)
        f_obj.write(output)
        f_obj.write("\n")



import datetime
import itertools
import json
import logging
import os.path
import tempfile
import unittest

import pkg_resources
from reportlab.platypus import SimpleDocTemplate
from reportlab.lib.pagesizes import A4

from sdx.common.survey import Survey
from sdx.common.test.test_transformer import PackingTests
from sdx.common.transforms.ImageTransformer import ImageTransformer
from sdx.common.transforms.PDFTransformer import PDFTransformer


class ImageTransformTests(unittest.TestCase):

    def setUp(self):
        self.max_diff, self.maxDiff = self.maxDiff, None

    def tearDown(self):
        self.maxDiff = self.max_diff

    def test_mwss_index(self):
        settings = PackingTests.Settings(
            "\\\\NP3RVWAPXX370\\SDX_preprod",
            "EDC_QImages"
        )
        log = logging.getLogger("")

        reply = json.loads(
            pkg_resources.resource_string(
                "sdx.common.test", "data/eq-mwss.json"
            ).decode("utf-8")
        )

        survey = json.loads(
            pkg_resources.resource_string(
                "sdx.common.test", "data/134.0005.json"
            ).decode("utf-8")
        )

        check = pkg_resources.resource_string(
            "sdx.common.test", "data/EDC_134_20170301_1000.csv"
        ).decode("utf-8")

        ids = Survey.identifiers(reply)._replace(
            ts=datetime.datetime(2017, 3, 7, 9, 45, 4)
        )

        with tempfile.TemporaryDirectory(prefix="sdx_") as locn:

            # Build PDF
            fp = os.path.join(locn, "pages.pdf")
            doc = SimpleDocTemplate(fp, pagesize=A4)
            doc.build(PDFTransformer.get_elements(survey, reply))

            # Create page images from PDF
            img_tfr = ImageTransformer(log, settings, survey, reply)
            images = list(img_tfr.create_image_sequence(fp, nmbr_seq=itertools.count()))

            self.assertEqual(
                check.splitlines(),
                list(img_tfr.index_lines(ids, images))
            )



#!/usr/bin/env python
#   coding: UTF-8

import configparser
import os.path
import re


def aggregate_options(cfg, name="default", keys=[]):
    """
    Create a keyword argument dictionary from the options in a configuration
    section. Override those values from the process environment.

    """
    if cfg.has_section(name):
        data = ((k, cfg.get(name, k)) for k in keys)
        rv = dict(pair for pair in data if pair[1] is not None)
    else:
        rv = {}

    variables = {
        k: "{0}_{1}".format(
            name.replace(".", "_").upper(), k.replace(".", "_").upper()
        )
        for k in keys
    }
    overlay = (
        (key, os.getenv(var))
        for key, var in variables.items()
    )
    rv.update(pair for pair in overlay if pair[1] is not None)
    return rv


def check_safe_value(val):
    r = re.compile("[$#]")
    return isinstance(val, str) and r.search(val) is None


def config_parser(content=None):
    rv = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation()
    )
    if content is not None:
        rv.read_string(content)
    return rv



#!/usr/bin/env python
# encoding: UTF-8

import ast
import os.path

from setuptools import setup

try:
    # For setup.py install
    from sdx.common import __version__ as version
except ImportError:
    # For pip installations
    version = str(
        ast.literal_eval(
            open(os.path.join(
                os.path.dirname(__file__),
                "sdx", "common", "__init__.py"),
                'r').read().split("=")[-1].strip()
        )
    )

__doc__ = open(os.path.join(os.path.dirname(__file__), "README.rst"),
               'r').read()

installRequirements = [
    i.strip() for i in open(
        os.path.join(os.path.dirname(__file__), "requirements.txt"), 'r'
    ).readlines()
]

setup(
    name="sdx-common",
    version=version,
    description="A shared library for SDX services",
    author="D Haynes",
    author_email="tundish@thuswise.org",
    url="https://github.com/ONSdigital/sdx-common",
    long_description=__doc__,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=[
        "sdx.common",
        "sdx.common.test",
        "sdx.common.formats",
        "sdx.common.formats.test",
        "sdx.common.transforms",
        "sdx.common.transforms.test",
    ],
    package_data={
        "sdx.common": [
            "requirements.txt",
            "doc/*.rst",
            "doc/_templates/*.css",
            "doc/html/*.html",
            "doc/html/*.js",
            "doc/html/_sources/*",
            "doc/html/_static/css/*",
            "doc/html/_static/font/*",
            "doc/html/_static/js/*",
            "doc/html/_static/*.css",
            "doc/html/_static/*.gif",
            "doc/html/_static/*.js",
            "doc/html/_static/*.png",
        ],
        "sdx.common.test": [
            "data/*.json",
        ],
    },
    install_requires=installRequirements,
    extras_require={
        "dev": [
            "pep8>=1.6.2",
            "flake8>=2.6.0"
        ],
        "docbuild": [
            "sphinx-argparse>=0.1.17",
            "sphinxcontrib-seqdiag>=0.8.5",
            "sphinx_rtd_theme>=0.2.4",
        ],
    },
    tests_require=[
    ],
    entry_points={
        "console_scripts": [
        ],
    },
    zip_safe=False
)



from collections import OrderedDict
import io
import logging
import os
import tempfile
import zipfile

from reportlab.platypus import SimpleDocTemplate
from reportlab.lib.pagesizes import A4
from structlog import wrap_logger

from sdx.common.formats.cs_formatter import CSFormatter
from sdx.common.survey import Survey
from sdx.common.transforms.ImageTransformer import ImageTransformer
from sdx.common.transforms.PDFTransformer import PDFTransformer

__doc__ = "Transform survey data into formats required downstream."


class Transformer:

    """A base class for SDX transformers.

    Subclasses should define the contents of the following class variables:

        * :py:const:`defn`
        * :py:const:`package`
        * :py:const:`pattern`

    """
    #: Transformer subclasses declare their transforms in this class variable.
    #: Each element is a 3-tuple consisting of:
    #:
    #: #. An integer or range corresponding to one or more question ids.
    #: #. A default value for the question(s).
    #: #. A :ref:`Processing function <processors>`.
    #:
    #: Eg, to declare question ids 151, 152, 153 as unsigned integers with a default
    #: of 0::
    #:
    #:  defn = [
    #:   (range(151, 154, 1), 0, Processor.unsigned_integer),
    #:
    #:  ]
    #:
    defn = []

    #: Defines the package where survey definitions can be found.
    #: For deployed Python packages, this will be a dotted package name.
    #: If your Transformer class is deployed from a source repository,
    #: you should set this class variable to `__name__`.
    package = "sdx.common"

    #: Defines a search pattern for survey definitions based on identifiers.
    #: The path is relative to the location specified by :py:const:`package` above.
    pattern = "../surveys/{survey_id}.{inst_id}.json"

    @classmethod
    def ops(cls):
        """Publish the sequence of operations for the transform.

        Return an ordered mapping from question id to default value and processing function.

        """
        return OrderedDict([
            ("{0:02}".format(qNr), (dflt, fn))
            for rng, dflt, fn in cls.defn
            for qNr in (rng if isinstance(rng, range) else [rng])
        ])

    @classmethod
    def transform(cls, data, survey=None):
        """Perform a transform on survey data."""
        return OrderedDict(
            (qid, fn(qid, data, dflt, survey))
            for qid, (dflt, fn) in cls.ops().items()
        )

    @staticmethod
    def create_zip(locn, manifest):
        """Create a zip archive from a local directory and a manifest list.

        Return the contents of the zip as bytes.

        """
        zip_bytes = io.BytesIO()

        with zipfile.ZipFile(zip_bytes, "w", zipfile.ZIP_DEFLATED) as zip_obj:
            for dst, f_name in manifest:
                zip_obj.write(os.path.join(locn, f_name), arcname=os.path.join(dst, f_name))

        zip_bytes.seek(0)
        return zip_bytes

    def __init__(self, response, seq_nr=0, log=None):
        """Create a transformer object to process a survey response."""
        self.response = response
        self.ids = Survey.identifiers(response, seq_nr=seq_nr)

        if self.ids is None:
            raise UserWarning("Missing identifiers")

        if log is None:
            self.log = wrap_logger(logging.getLogger(__name__))
        else:
            self.log = Survey.bind_logger(log, self.ids)

        for attr in ("defn", "package", "pattern"):
            if not hasattr(self.__class__, attr):
                raise UserWarning("Missing class attribute: {0}".format(attr))

    def pack(self, settings=None, img_seq=None, tmp="tmp"):
        """Perform transformation on the survey data and pack the output into a zip file.

        Return the contents of the zip as bytes.
        The object maintains a temporary directory while the output is generated.

        """
        survey = Survey.load_survey(self.ids, self.package, self.pattern)
        manifest = []
        with tempfile.TemporaryDirectory(prefix="sdx_", dir=tmp) as locn:
            # Do transform and write PCK
            data = self.transform(self.response["data"], survey)
            f_name = CSFormatter.pck_name(**self.ids._asdict())
            with open(os.path.join(locn, f_name), "w") as pck:
                CSFormatter.write_pck(pck, data, **self.ids._asdict())
            manifest.append(("EDC_QData", f_name))

            # Create IDBR file
            f_name = CSFormatter.idbr_name(**self.ids._asdict())
            with open(os.path.join(locn, f_name), "w") as idbr:
                CSFormatter.write_idbr(idbr, **self.ids._asdict())
            manifest.append(("EDC_QReceipts", f_name))

            # Build PDF
            fp = os.path.join(locn, "pages.pdf")
            doc = SimpleDocTemplate(fp, pagesize=A4)
            doc.build(PDFTransformer.get_elements(survey, self.response))

            # Create page images from PDF
            img_tfr = ImageTransformer(
                self.log, settings, survey, self.response, self.ids.seq_nr
            )
            images = list(img_tfr.create_image_sequence(fp, nmbr_seq=img_seq))
            for img in images:
                f_name = os.path.basename(img)
                manifest.append(("EDC_QImages/Images", f_name))

            # Write image index
            index = img_tfr.create_image_index(images)
            if index is not None:
                f_name = os.path.basename(index)
                manifest.append(("EDC_QImages/Index", f_name))

            return self.create_zip(locn, manifest)



from decimal import Decimal, InvalidOperation
from functools import reduce
import operator

from sdx.common.survey import Survey


__doc__ = """
The processor module collects business logic processing functions under a single
namespace so they can be used in :ref:`transformers`.

"""


class Processor:
    """Business logic operations on data.

    These methods are used to perform business logic on survey data.
    They are mostly concerned with combining multiple fields into a
    single field for output.

    Principles for processor methods:

    * The method is responsible for range check according to its own logic.
    * Parametrisation is possible; use `functools.partial` to bind arguments.
    * Return data of the same type as the supplied default.
    * On any error, return the default.

    """

    @staticmethod
    def aggregate(qid, data, default, *args, weights=[], **kwargs):
        """Calculate the weighted sum of a question group.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.
        :param weights: A sequence of 2-tuples giving the weight value for each
            question in the group.
        :type weights: [(str, number)]

        """
        try:
            return type(default)(
                Decimal(data.get(qid, 0)) +
                sum(Decimal(scale) * Decimal(data.get(q, 0)) for q, scale in weights)
            )
        except (InvalidOperation, ValueError):
            return default

    @staticmethod
    def evaluate(qid, data, default, *args, group=[], convert=bool, op=operator.or_, **kwargs):
        """Perform a map/reduce evaluation of a question group.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.
        :param group: A sequence of question ids.
        :param convert: A type or function to convert the group values.
        :param op: A binary operator or function to reduce data to a single value.

        """
        try:
            group_vals = [data.get(qid, None)] + [data.get(q, None) for q in group]
            data = [convert(i) for i in group_vals if i is not None]
            return type(default)(reduce(op, data))
        except (TypeError, ValueError):
            return default

    @staticmethod
    def mean(qid, data, default, *args, group=[], **kwargs):
        """Calculate the mean of all fields in a question group.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.
        :param group: A sequence of question ids.

        """
        try:
            group_vals = [data.get(qid, None)] + [data.get(q, None) for q in group]
            data = [Decimal(i) for i in group_vals if i is not None]
            divisor = len(data) or 1
            rv = sum(data) / divisor
            return type(default)(rv)
        except (AttributeError, InvalidOperation, TypeError, ValueError):
            return default

    @staticmethod
    def events(qid, data, default, *args, group=[], **kwargs):
        """Return a sequence of time events from a question group.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.
        :param group: A sequence of question ids.

        """
        try:
            group_vals = [data.get(qid, None)] + [data.get(q, None) for q in group]
            data = sorted(filter(
                None, (Survey.parse_timestamp(i) for i in group_vals if i is not None)
            ))
            if all(isinstance(i, type(default)) for i in data):
                return data
            else:
                return type(default)(data)
        except (AttributeError, TypeError, ValueError):
            return default

    @staticmethod
    def survey_string(qid, data, default, *args, survey=None, **kwargs):
        """Accept a string as an option for a question.

        This method provides an opportunity for validating the string against
        the survey definition, though this has not been a requirement so far.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.
        :param dict survey: The survey definition.

        """
        try:
            return type(default)(data[qid])
        except (KeyError, ValueError):
            return default

    @staticmethod
    def unsigned_integer(qid, data, default, *args, **kwargs):
        """Process a string as an unsigned integer.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.

        """
        try:
            rv = int(data.get(qid, default))
        except ValueError:
            return default
        else:
            return type(default)(rv) if rv >= 0 else default

    @staticmethod
    def percentage(qid, data, default, *args, **kwargs):
        """Process a string as a number, checking that it is valid as a percentage.

        :param str qid: The question id.
        :param data: The full survey data.
        :type data: dict(str, str)
        :param default: The default value for the question.

        """
        try:
            rv = Decimal(data.get(qid, default))
        except ValueError:
            return default
        else:
            return type(default)(rv) if 0 <= rv <= 100 else default
